// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of scatter
//
// scatter is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// scatter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include  "MZ.h"

#include <math.h>
#include <mathconst/mathconst.h>

#include <matvec/macros.h>

#define ARCSEC2RAD(theta) ( (theta / 3600) * (M_PI / 180) )
#define RAD2ARCSEC(theta) ( (theta / M_PI) * 180 * 3600 )


// use matvec macros, but simplify them

#define VEC(vec) MATVEC_VEC_AS_STRUCT(vec,x,y,z)

#define VCROSS(p, v1, v2)	\
  MATVEC_VEC_CROSSPROD( VEC(p), VEC(v1), VEC(v2) )

#define VDOT(v1, v2)		\
  MATVEC_VEC_DOTPROD( VEC(v1), VEC(v2))

#define VLEN(v) sqrt(VDOT((v),(v)))

#define VADD( v, v1, v2 )	\
  MATVEC_VEC_ADD( VEC(v), VEC(v1), VEC(v2) )

#define VADDTO( v1, v2 )	\
  MATVEC_VEC_ADD( VEC(v1), VEC(v1), VEC(v2) )

#define VSET( v1, v2 )	\
  MATVEC_VEC_ASSIGN( VEC(v1), VEC(v2 ) )

#define VSCALE( r, v, c )	\
  MATVEC_VEC_SCALE( VEC(r), VEC(v), c )

#define VNORM(v)	\
  MATVEC_VEC_NORM(VEC(v))


MZ::MZ( const string& id ) throw ( Exception ) : _id(id), _zmin(DBL_MIN), _zmax(DBL_MAX) {

  TraceFct tf( "MZ::MZ" );

  _data = NULL;

}

MZ::~MZ( void ) {

  TraceFct tf( "MZ::~MZ" );

  if ( _data )
    free( _data );

}

int
MZ::scatter( 
  ParFile& pf,
  const long id,
  const double energy,
  const DVector3& surface_norm,
  DVector3& direction,
  double& theta_o,
  double& phi_o
)
{
  TraceFct tf( "MZ::scatter" );

  double sin_alpha  = -1.0 * VDOT( surface_norm, direction );
  double cos_alpha  = sqrt( 1.0 - sin_alpha * sin_alpha );

  // default scatter angles; convert from arcseconds to radians
  double theta      = ARCSEC2RAD(pf.pgetd( "theta" ));
  double phi        = ARCSEC2RAD(pf.pgetd( "phi" ));

  double delta_prob = pf.pgetd( "max_cumulative_probability" ) - pf.pgetd( "min_cumulative_probability" );
  double min_prob   = pf.pgetd( "min_cumulative_probability" );

  int iter = pf.pgeti( "scat_iter_lim" );

  int status = OK;


#ifdef DEBUG
  double alpha	    = asin( sin_alpha );
#endif // DEBUG

#ifdef DEBUG
  if ( dbflag( "zones" ) ) 
    tf.message( "zones:\t%u\t%s\n", id, _id.c_str( ) );
#endif // DEBUG
    
  // temporary direction vector to handle possible iterations
  DVector3 tdir;

  do
  {
    // so far, nothing bad has happened
    status = OK;

    tdir = direction;

    if ( pf.pgetb( "in_plane" ) && 
	 lookUpScatterAngle( energy, sin_alpha, min_prob + ranf( ) * delta_prob, true, theta ) )
      status |= IP_EXTRAPOLATED;

    if ( pf.pgetb( "out_of_plane" ) && 
	 lookUpScatterAngle( energy, sin_alpha, min_prob + ranf( ) * delta_prob, false, phi ) )
      status |= OP_EXTRAPOLATED;

    theta_o  = theta;
    phi_o = phi;

    double cos_theta = cos( theta );
    double sin_theta = sin( theta );
    double cos_phi   = cos( phi   );
    double sin_phi   = sin( phi   );
  
    // cos(alpha+theta)
    double cos_at = cos_alpha * cos_theta - sin_alpha * sin_theta;
    // sin(alpha+theta)
    double sin_at = sin_alpha * cos_theta + cos_alpha * sin_theta;
  
    double A = cos_at / cos_alpha;
    double B = A * sin_alpha * cos_phi - sin_at;
  
    DVector3 r_cross_n;
    VCROSS(r_cross_n, direction, surface_norm );

    DVector3 v;
    VSCALE(v, surface_norm, sin_alpha);
    VADDTO(v, direction);
    VSCALE(v, v, 1/cos_alpha );

#ifdef DEBUG
    DVector3 u;

    VCROSS( u, direction, surface_norm );
    VNORM(u);

    //DVector3 v;
    //VCROSS( v, surface_norm, u );

#endif // DEBUG

  
    /* Construct new reflected ray. note that it's a unit vector */
  
    tdir.x = B * surface_norm.x + A * (cos_phi * direction.x + sin_phi * r_cross_n.x);
    tdir.y = B * surface_norm.y + A * (cos_phi * direction.y + sin_phi * r_cross_n.y);
    tdir.z = B * surface_norm.z + A * (cos_phi * direction.z + sin_phi * r_cross_n.z);
  
    // not quite normalized due to numerics. fix that.
    VNORM(tdir);

    // has the ray been scattered into the reflecting surface?
    if ( VDOT( tdir, surface_norm ) > 0 )
      status |= SCATTERED_INTO_SURFACE;

    // has the ray been backscattered? check that the v components
    // have the same sign (i.e. they're on the same side of the
    // surface normal in the specular plane.  if they have the same
    // sign their product will be >= 0;
    if ( VDOT( tdir, v) * VDOT( direction, v ) < 0 )
      status |= BACKSCATTERED;

#ifdef DEBUG

    if ( dbflag( "ray" ) )
      tf.message( "ray:\t(%.4f %.4f %.4f) -> (%.4f %.4f %.4f)\n", 
		  direction.x, direction.y, direction.z, tdir.x, tdir.y, tdir.z );

    if ( dbflag( "norm" ) )
      tf.message( "norm:\t(%.4f %.4f %.4f)\n", surface_norm.x, surface_norm.y, surface_norm.z );

    if ( dbflag( "theta" ) ) 
    {
      double theta_out = asin( - VDOT(tdir, surface_norm ) ) - alpha;
      tf.message( "theta_in:\t%u\t%.14g\n", id, RAD2ARCSEC(theta) );
      tf.message( "theta_out:\t%u\t%.14g\n", id, RAD2ARCSEC(theta_out) );
    }

    if ( dbflag( "phi" ) ) 
    {
      DVector3 _v;
      DVector3 _u;

      VSCALE( _v, v, VDOT(tdir, v) );

      VSCALE( _u, u, VDOT(tdir, u) );

      DVector3 l;

      VADD( l, _v, _u );

      VNORM(l);

      tf.message( "phi_in:\t\t%u\t%.14g\n", id, RAD2ARCSEC(phi) );
      tf.message( "phi_out:\t%u\t%.14g\n", id, RAD2ARCSEC(acos( VDOT(l, v ) ) * ( VDOT(l,u) < 0 ? -1 : 1 )) );
    }

#endif // DEBUG


  } while( --iter && status & SCATTERED_INTO_SURFACE );

  direction = tdir;

  return status;
}


double MZ::interpolate( double x, double x1, double y1, double x2, double y2 ) {
  TraceFct tf( "MZ::interpolate" );

  double y;
  if ( x1 == x2 ) 
    y = y1;
  else {
    double f2 = ( x - x1 ) / ( x2 - x1 );
    double f1 = 1 - f2;
    y  =  f1 * y1 + f2 * y2;
  }

  return y;
}


void MZ::verifyCards( Fun fun ) throw ( Exception ) {

  TraceFct tf( "MZ::verifyCards" );

  FunInfoGet( fun, FUN_ROWSIZE, &_rowsize, 0 );

  int got = 0;
  if ( 0 > (_nrows = FunParamGeti( fun, const_cast<char*>("NAXIS"), 2, -1, &got )) || ! got ) 
    throw( Exception( "error determining number of rows" ) );

}

void MZ::verifyColumns( Fun fun, vector<string>& names ) throw ( Exception ) {

  TraceFct tf( "MZ::verifyColumns" );

  for ( vector<string>::iterator itr = names.begin( ); names.end( ) != itr; itr++ ) {

    int type, offset, n;
    if ( ! FunColumnLookup( fun, const_cast<char*>(itr->c_str( )), 0, NULL, &type, NULL, &offset, &n, NULL ) )
      throw( Exception( string("error finding column '") + *itr + "'" ) );

    if ( 'D' != type ) 
      throw( Exception( string("invalid data type for column '") + *itr + "'" ) );

    _offsets.insert( make_pair( *itr, offset ) );
    _ns.insert(      make_pair( *itr, n      ) );

  }

}

void MZ::verifyData( Fun fun ) throw ( Exception ) {

  TraceFct tf( "MZ::verifyData" );

  int nread = 0;
  if ( ! (_data = (char *)FunTableRowGet( fun, NULL, _nrows, NULL, &nread )) )
    throw( Exception( "error reading data" ) );

}
